package adaptivehac;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Vector;

abstract public class AdaptiveHAC implements Serializable
{
  private static final long serialVersionUID = 1L;
  protected RankingModel rm_;
  
  public RankingModel getRankingModel()
  {
    return rm_;
  }
  
  abstract public void train(Document[] docs);

  public Pair<? extends Cluster> getEmptyPair()
  {
    return Cluster.EMPTY_PAIR_;
  }
  
  public Clustering[] cluster(Document[] docs)
  {
    Clustering[] clustering = new Clustering[docs.length];
    for (int i = 0; i < docs.length; i++) {
      clustering[i] = cluster(docs[i]);
      System.out.println("Done document " + i);
    }
    
    return clustering;
  }
  
  public Clustering cluster(Document doc)
  {
    Clustering clustering = doc.getSingletonClusters();
    clustering.setFeatureVectors(doc, rm_.alphabet_);
    Pair<Cluster> best_pair;
    do {
      best_pair = getBestPair(doc, clustering);
      if (best_pair != getEmptyPair()) {
        Cluster c = clustering.join(best_pair.getLeft(), best_pair.getRight());
        clustering.updateFeatureVectors(doc, rm_.alphabet_, c);
      }
    } 
    while (best_pair != getEmptyPair());
    
    return clustering;
  }


  public Clustering[] clusterPWSoon(Document[] docs)
  {
    double empty_score = rm_.evaluate(docs[0], getEmptyPair());
    System.out.println("Empty score = " +  empty_score);
    
    Clustering[] clustering = new Clustering[docs.length];
    for (int i = 0; i < docs.length; i++) {
      clustering[i] = clusterPWSoon(docs[i]);
      System.out.println("Done document " + i);
    }
    
    return clustering;
  }
  

  @SuppressWarnings("unchecked")
  public Clustering clusterPWSoon(Document doc)
  {    
    Clustering singletons = doc.getSingletonClusters();
    int[] grounding = new int[singletons.size()];
    
    double empty_score = rm_.evaluate(doc, getEmptyPair());
    int chains = 0;
    
    // Assume singleton clusters are ordered
    for (int i = 0; i < singletons.size(); i++) {
      Cluster anaphor = singletons.get(i);
      
      // Find best scoring antecedent.
      Pair<? extends Cluster> best_pair = getEmptyPair();
      double best_score = empty_score;
      for (int j = i - 1; j >=0; j--) {
        Cluster antecedent = singletons.get(j);
        Pair<Cluster> pair = Pair.create(antecedent, anaphor);
        double score = rm_.evaluate(doc, pair);
        if (score > best_score) {
          best_pair = pair;
          best_score = score;
          grounding[i] = grounding[j];
//          break;
        }
      }
      

      if (best_pair == getEmptyPair())
        grounding[i] = chains++;
    }
    
    // Create final clusters.
    Vector<Element>[] elements = new Vector[chains];
    for (int i = 0; i < elements.length; i++)
      elements[i] = new Vector<Element>();
    for (int i = 0; i < grounding.length; i++)
      elements[grounding[i]].add(singletons.get(i).get(0));
    
    // Create final clustering. Change this to make decisions in decreasing 
    // order of confidence, to allow computing PR curves.
    Clustering clustering = new Clustering();
    for (int i = 0; i < elements.length; i++)
      clustering.add(new Cluster(elements[i].toArray((new Element[0])), null, null));
    
    return clustering;
  }



  static class TauElement implements Comparable<TauElement>
  {
    Document doc_;
    Clustering chains_;
    double tau_;
    int system_;
    int common_;
    int last_system_;
    int last_common_;
    
    public TauElement(Document doc)
    {
      doc_ = doc;
      chains_ = doc.getSingletonClusters();
      tau_ = Double.POSITIVE_INFINITY;
      common_ = system_ = doc.elements_.length;
    }
    
    public int compareTo(TauElement te)
    {
      if (tau_ < te.tau_)
        return 1;
      if (tau_ > te.tau_)
        return -1;
      return 0;
    }
    
    public void add(PRElement pre)
    {
      system_ += pre.system_;
      common_ += pre.common_;
    }
  }
  
  static public class PRElement
  {
    public double tau_;
    public int system_;
    public int common_;
    
    public PRElement()
    {
      tau_ = Double.POSITIVE_INFINITY;
      system_ = common_ = 0;
    }
    
    public PRElement(double tau, int system, int common)
    {
      tau_ = tau;
      system_ = system;
      common_ = common;
    }
    
    public PRElement(TauElement te)
    {
      tau_ = te.tau_;
      system_ = te.system_;
      common_ = te.common_;
    }
    
    public PRElement(PRElement te)
    {
      tau_ = te.tau_;
      system_ = te.system_;
      common_ = te.common_;
    }
  }

  public double cluster(TauElement te)
  {
    double tau;
    if (te.chains_.size() == 1)
      tau = Double.NEGATIVE_INFINITY;
    else {
      Pair<Cluster> best_pair = getBestNonEmptyPair(te.doc_, te.chains_);
      tau = rm_.evaluate(te.doc_, best_pair);

      Cluster l = best_pair.getLeft();
      Cluster r = best_pair.getRight();
      te.last_system_ = 2 * l.size() * r.size();
      te.system_ += te.last_system_;

      Cluster lr = te.chains_.join(l, r);
      int common_l = getB3Common(l);
      int common_r = getB3Common(r);
      int common_lr = getB3Common(lr);
      te.last_common_ = common_lr - common_l -common_r;
      te.common_ += te.last_common_;
    }
    
    return tau;
  }
  
  public int getB3Common(Cluster cluster)
  {
    int common = 0;
    for (int i = 0; i < cluster.elements_.length; i++) {
      Element el = cluster.elements_[i];
      for (int j = 0; j < cluster.elements_.length; j++) {
        Element ela = cluster.elements_[j];
        if (ela.getClusterID() == el.getClusterID())
          common++;
      }
    }

    return common;
  }
  
  public PRElement[] clusterPRCurve(Document[] docs)
  {
    Vector<PRElement> pr_elems = new Vector<PRElement>();
    
    TauElement[] tau_elems = new TauElement[docs.length];
    for (int i = 0; i < docs.length; i++) {
      tau_elems[i] = new TauElement(docs[i]);
    }
    
    // Initial clustering.
    for (TauElement tau_elem : tau_elems) {
      tau_elem.tau_ = cluster(tau_elem);
      tau_elem.last_system_ = tau_elem.system_;
      tau_elem.last_common_ = tau_elem.common_;
    }
    Arrays.sort(tau_elems);
    
    PRElement pr_total = new PRElement();
    while (tau_elems[0].tau_ > Double.NEGATIVE_INFINITY) {
      // Cluster until tau gets lower.
      pr_total.tau_ = tau_elems[0].tau_;
      pr_total.common_ += tau_elems[0].last_common_;
      pr_total.system_ += tau_elems[0].last_system_;
      double new_tau = cluster(tau_elems[0]);
      while (new_tau >= tau_elems[0].tau_) {
        pr_total.common_ += tau_elems[0].last_common_;
        pr_total.system_ += tau_elems[0].last_system_;
        new_tau = cluster(tau_elems[0]);
      }
      tau_elems[0].tau_ = new_tau;
      pr_elems.add(new PRElement(pr_total));
      
      if (pr_elems.size() % 100 == 0)
        System.out.println(pr_elems.size());
      
      // Insertion sort tau elements.
      TauElement key = tau_elems[0];
      int i = 1;
      for (; i < tau_elems.length; i++) {
        if (tau_elems[i].tau_<= key.tau_)
          break;
        tau_elems[i - 1] = tau_elems[i];
      }
      tau_elems[i - 1] = key;
    }
    
    return pr_elems.toArray(new PRElement[0]);
  }
  
  
  static class TauPWElement implements Comparable<TauPWElement>
  {
    static class LinkDecision implements Comparable<LinkDecision>
    {
      int elem_;
      int ante_;
      double score_;
      
      public LinkDecision(int elem, int ante, double score)
      {
        elem_ = elem;
        ante_ = ante;
        score_ = score;
      }
      
      public int compareTo(LinkDecision ld)
      {
        if (score_ < ld.score_)
          return 1;
        if (score_ > ld.score_)
          return -1;
        return 0;
      }
    }
    
    Document doc_;
    Clustering chains_;
    Cluster[] el_to_chain_;
    LinkDecision[] ld_;
    int crt_;
    
    int system_;
    int common_;
    int last_system_;
    int last_common_;
    
    public TauPWElement(Document doc, RankingModel rm)
    {
      doc_ = doc;
      chains_ = doc.getSingletonClusters();
      
      el_to_chain_ = new Cluster[doc_.elements_.length];
      for (int i = 0; i < el_to_chain_.length; i++)
        el_to_chain_[i] = chains_.get(i);
      
//      double empty_score = rm.evaluate(doc, Cluster.EMPTY_PAIR_);
      ld_ = new LinkDecision[doc_.elements_.length];     
      // Assume singleton clusters are ordered
      for (int i = 0; i < chains_.size(); i++) {
        Cluster anaphor = chains_.get(i);       
        // Find best scoring antecedent.
        int best_ante = -1;
        double best_score = Double.NEGATIVE_INFINITY;
        for (int j = i - 1; j >=0; j--) {
          Cluster antecedent = chains_.get(j);
          Pair<Cluster> pair = Pair.create(antecedent, anaphor);
          double score = rm.evaluate(doc, pair);
          if (score > best_score) {
            best_ante = j;
            best_score = score;
          }
//          if (score > empty_score)
//            break;
        }
        ld_[i] = new LinkDecision(i, best_ante, best_score);
      }
      Arrays.sort(ld_);
      
      crt_ = 0;
      common_ = system_ = doc.elements_.length;
      last_common_ = common_;
      last_system_ = system_;
    }
    
    public double getScore()
    {
      return ld_[crt_].score_;
    }
    
    public int compareTo(TauPWElement te)
    {
      if (getScore() < te.getScore())
        return 1;
      if (getScore() > te.getScore())
        return -1;
      return 0;
    }
    
    public void add(PRElement pre)
    {
      system_ += pre.system_;
      common_ += pre.common_;
    }
  }
  

  public PRElement[] clusterPWPRCurve(Document[] docs)
  {
    Vector<PRElement> pr_elems = new Vector<PRElement>();
    
    PRElement pr_total = new PRElement();
    TauPWElement[] tau_elems = new TauPWElement[docs.length];
    for (int i = 0; i < docs.length; i++) {
      tau_elems[i] = new TauPWElement(docs[i], rm_);
      pr_total.common_ += tau_elems[i].last_common_;
      pr_total.system_ += tau_elems[i].last_system_;
    }
    Arrays.sort(tau_elems);
    pr_total.tau_ = Double.POSITIVE_INFINITY;
    pr_elems.add(new PRElement(pr_total));
  
    while (tau_elems[0].getScore() > Double.NEGATIVE_INFINITY) {
      pr_total.tau_ = cluster(tau_elems[0]);
      pr_total.common_ += tau_elems[0].last_common_;
      pr_total.system_ += tau_elems[0].last_system_;
      pr_elems.add(new PRElement(pr_total));

      if (pr_elems.size() % 100 == 0)
        System.out.println(pr_elems.size());
      
      // Insertion sort tau elements.
      TauPWElement key = tau_elems[0];
      int i = 1;
      for (; i < tau_elems.length; i++) {
        if (tau_elems[i].getScore() <= key.getScore())
          break;
        tau_elems[i - 1] = tau_elems[i];
      }
      tau_elems[i - 1] = key;
    }
    
    return pr_elems.toArray(new PRElement[0]);
  }
  
  
  public double cluster(TauPWElement te)
  {
    double result = te.getScore();
    
    Cluster l = te.el_to_chain_[te.ld_[te.crt_].ante_];
    Cluster r = te.el_to_chain_[te.ld_[te.crt_].elem_];
    te.last_system_ = 2 * l.size() * r.size();
    te.system_ += te.last_system_;

    Cluster lr = te.chains_.join(l, r);
    int common_l = getB3Common(l);
    int common_r = getB3Common(r);
    int common_lr = getB3Common(lr);
    te.last_common_ = common_lr - common_l -common_r;
    te.common_ += te.last_common_;
    for (Element el : lr.elements_)
      te.el_to_chain_[el.index_] = lr;
    
    te.crt_++;
    
    return result;
  }
  

  public int getB3True(Document[] docs)
  {
    int total_true = 0;
    for (Document doc : docs)
      total_true += getB3True(doc);
    
    return total_true;
  }
  
  public int getB3True(Document doc)
  {
    int total_true = 0;
    Clustering cl = doc.getTrueClustering();
    for (Cluster c : cl.clusters_)
      total_true += c.size() * c.size();

    return total_true;
  }
  
  public Pair<Cluster> getBestPair(Document doc, Clustering clustering)
  {
    Pair<Cluster> best_pair = Cluster.EMPTY_PAIR_;
    double best_score = rm_.evaluate(doc, best_pair);
    
    for (int i = 0; i < clustering.size(); i++) {
      for (int j = i + 1; j < clustering.size(); j++) {
        Pair<Cluster> pair = Pair.create(clustering.get(i), clustering.get(j));
        double score = rm_.evaluate(doc, pair);
        if (score > best_score) {
          best_pair = pair;
          best_score = score;
        }
      }
    }
    
    return best_pair;
  }
  
 
  public Pair<Cluster> getBestNonEmptyPair(Document doc, Clustering clustering)
  {
    Pair<Cluster> best_pair = null;
    double best_score = Double.NEGATIVE_INFINITY;
    
    for (int i = 0; i < clustering.size(); i++) {
      for (int j = i + 1; j < clustering.size(); j++) {
        Pair<Cluster> pair = Pair.create(clustering.get(i), clustering.get(j));
        double score = rm_.evaluate(doc, pair);
        if (score > best_score) {
          best_pair = pair;
          best_score = score;
        }
      }
    }
    
    return best_pair;
  }
 
}
