package adaptivehac;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Vector;
import java.util.Map.Entry;

public abstract class Document
{
  // True if document generates cluster-wise features.
  static public boolean CLUSTER_WISE_ = false;
  
  // A unique document identifier.
  public int id_;
  
  // The elements contained in this document;
  public Element[] elements_;
  
  // Upper triangular matrix of cached feature vectors for element pairs.
  transient TriangularMatrix<FeatureVector> fvs_;

  public Document()
  {
    id_ = -1;
    elements_ = null;
  }
  
  public Document(int id, Element[] elements)
  {
    id_ = id;
    elements_ = elements;
    fvs_ = new TriangularMatrix<FeatureVector>(elements_.length);
    
    // Sort and set elements' indices.
    Arrays.sort(elements);
    for (int i = 0; i < elements_.length; i++)
      elements_[i].index_ = i;
  }
  
  public int size()
  {
    return elements_.length;
  }
  
  public void clearFeatureVectors()
  {
    fvs_.clear();
  }
  
  public FeatureVector getFeatureVector(Element el1, Element el2)
  {
    // Assume el1 < el2
    int i1 = el1.index_;
    int i2 = el2.index_;
    
    return fvs_.get(i1, i2);
  }
  
  public void setFeatureVector(Element el1, Element el2, FeatureVector fv)
  {
    // Assume el1 < el2
    int i1 = el1.index_;
    int i2 = el2.index_;

    fvs_.set(i1, i2, fv);
  }
  
  public Clustering getSingletonClusters()
  {
    Clustering clustering = new Clustering();
    
    for (Element element : elements_) {
      Cluster cluster = new Cluster(element);
      clustering.add(cluster);
    }
    
    return clustering;
  }

  public Clustering getTrueClustering()
  {
    TreeMap<String, Vector<Element>> clusters = new TreeMap<String, Vector<Element>>();
    for (Element elem : elements_) {
      Vector<Element> elems = clusters.get(elem.getClusterID());
      if (elems == null) {
        elems = new Vector<Element>();
        clusters.put(elem.getClusterID().toString(), elems);
      }
      elems.add(elem);
    }
    
    Clustering result = new Clustering();
    Iterator<Vector<Element>> it = clusters.values().iterator();
    while (it.hasNext()) {
      Cluster cluster = new Cluster(it.next().toArray(new Element[0]), null, null);
      result.add(cluster);
    }
    
    return result;
  }
  
  protected FeatureVector getFeatureConjunctions(FeatureVector fv, AlphabetHAC alphabet)
  {
    // Create mapping index to feature strings.
    Vector<String> features = alphabet.tokens; 
    /*
    String[] features = new String[alphabet.size() + 1];
    Iterator<Entry<String, Integer>> it = alphabet.map.entrySet().iterator();
    while (it.hasNext()) {
      Entry<String, Integer> entry = it.next();
      features[entry.getValue()] = entry.getKey();
    }
    */ 
    
    // Create conjunctions of features.
    FeatureVector conj = new FeatureVector();
    for (int i = 0; i <  fv.features_.size(); i++) {
      Feature f1 = fv.features_.get(i);
      for (int j = i + 1; j <  fv.features_.size(); j++) {
        Feature f2 = fv.features_.get(j);
        String f;
        if (features.get(f1.index_).compareTo(features.get(f2.index_)) <= 0)
          f = features.get(f1.index_) + " & " + features.get(f2.index_);
        else
          f = features.get(f2.index_) + " & " + features.get(f1.index_);
        Integer intf = alphabet.getI(f);
        if (intf != null)
          conj.add(intf, f1.value_ * f2.value_);
      }
    }
    
    return conj;
  }
  
  
  protected FeatureVector getSumFeatureVector(FeatureVector fv, AlphabetHAC alphabet)
  {
    HashMap<Integer, Double> sum = new HashMap<Integer, Double>();
    for (int i = 0; i < fv.features_.size(); i++) {
      Feature feature = fv.features_.get(i);
      Double val = sum.get(feature.index_);
      if (val == null)
        val = new Double(0);
      sum.put(feature.index_, val + feature.value_);
    }
    
    FeatureVector bfv = new FeatureVector();
    Iterator<Entry<Integer, Double>> it = sum.entrySet().iterator();
    while (it.hasNext()) {
      Entry<Integer, Double> ii = it.next();
      /*
      String ftext = "sum" + "/" + alphabet.tokens.get(ii.getKey());
      Integer fidx = alphabet.get(ftext);
      if (fidx != null)
        bfv.add(fidx, ii.getValue());
      */
      bfv.add(ii.getKey(), ii.getValue());
    }
    
    return bfv;
  }
  
  abstract public Vector<? extends Pair<? extends Pair<? extends Cluster>>> getInitialExamples();
  
  abstract public Vector<Pair<Pair<Cluster>>> getInitialNegativeExamples();
  
  abstract public FeatureVector getFeatureVector(Pair<? extends Cluster> pair,
                                                 AlphabetHAC alphabet);
  
  public boolean isAllowedPair(Cluster c1, Cluster c2)
  {
    return true;
  }
  
}
