package adaptivehac;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

public class MiraRankingModel extends RankingModel
{
  private static final long serialVersionUID = 1L;
  
  transient LinearDiscriminant avg_;
  transient int updates_;
  transient int rounds_;

  public MiraRankingModel()
  {
    super();
    disc_ = new LinearDiscriminant();
    avg_ = new LinearDiscriminant();
    updates_ = 0;
    rounds_ = 0;
  }
  
  public double evaluate(Document doc, Pair<? extends Cluster> pair)
  {
    FeatureVector fv = doc.getFeatureVector(pair, alphabet_);
    return disc_.evaluate(fv);
  }

  @SuppressWarnings("unchecked")
  public void update(Document doc, Pair<? extends Pair<? extends Cluster>> pair)
  {
    Pair<? extends Cluster> positive = pair.getLeft();
    FeatureVector pos_fv = doc.getFeatureVector(positive, alphabet_);
    Pair<? extends Cluster> negative = pair.getRight();
    FeatureVector neg_fv = doc.getFeatureVector(negative, alphabet_);
    
    int K = 1;
    double[] b = new double[K];
    double[] lam_dist = new double[K];
    Pair<FeatureVector>[] dist = new Pair[K];
    
    lam_dist[0] = disc_.evaluate(pos_fv) - disc_.evaluate(neg_fv);
    b[0] = (double) numErrors(positive, negative) - lam_dist[0];
    dist[0] = Pair.create(pos_fv, neg_fv); 

    double[] alpha = hildreth(dist,b);
    for(int k = 0; k < K; k++) {
      Pair<FeatureVector> fv = dist[k];
      ((LinearDiscriminant) disc_).add(fv.getLeft(), alpha[k]);
      ((LinearDiscriminant) disc_).sub(fv.getRight(), alpha[k]);
    }
    
    updates_++;
  }
  
  private int numErrors(Pair<? extends Cluster> positive, Pair<? extends Cluster> negative)
  {
    return 1;
  }
  
  private double[] hildreth(Pair<FeatureVector>[] a, double[] b) {

    int i;
    int max_iter = 10000;
    double eps = 0.00000001;
    double zero = 0.000000000001;
    
    double[] alpha = new double[b.length];

    double[] F = new double[b.length];
    double[] kkt = new double[b.length];
    double max_kkt = Double.NEGATIVE_INFINITY;

    int K = a.length;
    
    double[][] A = new double[K][K];
    boolean[] is_computed = new boolean[K];
    for(i = 0; i < K; i++) {
      A[i][i] = dotProduct(a[i], a[i]);
      is_computed[i] = false;
    }
        
    int max_kkt_i = -1;

    
    for(i = 0; i < F.length; i++) {
      F[i] = b[i];
      kkt[i] = F[i];
      if(kkt[i] > max_kkt) { max_kkt = kkt[i]; max_kkt_i = i; }
    }

    int iter = 0;
    double diff_alpha;
    double try_alpha;
    double add_alpha;
  
    while(max_kkt >= eps && iter < max_iter) {
      
      diff_alpha = A[max_kkt_i][max_kkt_i] <= zero ? 0.0 : F[max_kkt_i]/A[max_kkt_i][max_kkt_i];
      try_alpha = alpha[max_kkt_i] + diff_alpha;
      add_alpha = 0.0;

      if(try_alpha < 0.0)
        add_alpha = -1.0 * alpha[max_kkt_i];
      else
        add_alpha = diff_alpha;

      alpha[max_kkt_i] = alpha[max_kkt_i] + add_alpha;

      if (!is_computed[max_kkt_i]) {
        for(i = 0; i < K; i++) {
          A[i][max_kkt_i] = dotProduct(a[i], a[max_kkt_i]); // for version 1
          is_computed[max_kkt_i] = true;
        }
      }

      for(i = 0; i < F.length; i++) {
        F[i] -= add_alpha * A[i][max_kkt_i];
        kkt[i] = F[i];
        if(alpha[i] > zero)
          kkt[i] = Math.abs(F[i]);
      }

      max_kkt = Double.NEGATIVE_INFINITY;
      max_kkt_i = -1;
      for(i = 0; i < F.length; i++)
        if(kkt[i] > max_kkt) { max_kkt = kkt[i]; max_kkt_i = i; }

      iter++;
    }

    return alpha;
  }

    
  private double dotProduct(Pair<FeatureVector> p1, Pair<FeatureVector> p2)
  {
    HashMap<Integer, Double> m1 = new HashMap<Integer, Double>();
    addFeaturesToMap(m1, p1.getLeft(), false);
    addFeaturesToMap(m1, p1.getRight(), true);
    
    HashMap<Integer, Double> m2 = new HashMap<Integer, Double>();
    addFeaturesToMap(m2, p2.getLeft(), false);
    addFeaturesToMap(m2, p2.getRight(), true);
    
    double result = 0.0;
    Iterator<Map.Entry<Integer, Double>> it = m1.entrySet().iterator();
    while (it.hasNext()) {
      Map.Entry<Integer, Double> entry = it.next();
      Double value = m2.get(entry.getKey());
      if (value != null)
        result += entry.getValue() * value;
    }
    
    return result;
  }
 
  
  private void addFeaturesToMap(HashMap<Integer, Double> m, FeatureVector fv, boolean negative)
  {
    for (Feature f : fv.features_) {
      Double value = m.get(f.index_);
      if (value == null)
        value = new Double(0);
      value = negative ? value - f.value_ : value + f.value_;
      m.put(f.index_, value);
    }
  }
  
  
  public void updateAvg()
  {
    avg_.add((LinearDiscriminant) disc_);
    rounds_++;
  }
  
  public void setAverageDiscriminant()
  {
    avg_.scale(1.0 / rounds_);
    disc_ = avg_;
    avg_ = null;
  }
  
  public void printFeatureWeights()
  {
    LinearDiscriminant disc = (LinearDiscriminant) disc_;
    Iterator<Map.Entry<String, Integer>> it = alphabet_.map.entrySet().iterator();
    while (it.hasNext()) {
      Map.Entry<String, Integer> entry = it.next();
      System.out.println(entry.getValue() + " --> " + entry.getKey() + " = " + disc.w_.get(entry.getValue()));
    }
  }

}
