package adaptivehac;

import java.util.Vector;
import java.util.Random;

public class SampleSumHAC extends AdaptiveHAC {

  private static final long serialVersionUID = 1L;
  static public int NUM_EPOCHS_ = 10;
  static public int MAX_DISTANCE_ = Integer.MAX_VALUE;//50;
  static public double EPSILON_ = 1E-10;
  
  private Random random_ = new Random(0);
  
  public SampleSumHAC(PerceptronRankingModel rm)
  {
    rm_ = rm;
  }
  
  public SampleSumHAC(MiraRankingModel rm)
  {
    rm_ = rm;
  }
  
  public void trainMira(Document[] docs)
  {
    MiraRankingModel rm = (MiraRankingModel) rm_;
    
    for (int epoch = 0; epoch < NUM_EPOCHS_; epoch++) {
      System.out.println("Training epoch " + epoch + ".");
      for (Document doc : docs) {
        System.out.println("\tDocument " + doc.id_);
        updateModelMira(doc);
      }
    }

    rm.setAverageDiscriminant();
    rm.alphabet_.stopGrowth();
    rm.printFeatureWeights();
  }
  
  
  protected void updateModelMira(Document doc)
  {

    MiraRankingModel rm = (MiraRankingModel) rm_;
    
    Clustering clustering = doc.getSingletonClusters();
    clustering.setFeatureVectors(doc, rm.alphabet_, rm);
    
    int decisions = 0;
    
    DecisionInfo best_pair = getBestScoringPair(doc, clustering); // can be empty
    while (best_pair != null) {
      DecisionInfo better_pair = getBetterAccuracyPair(doc, clustering, best_pair); // can be empty

      // Higher accuracy pair should have higher score.
      if (better_pair != null)
        rm.update(doc, Pair.create(better_pair.getPair(), best_pair.getPair()));
      rm.updateAvg();
      
      if (best_pair.getPair() == Cluster.EMPTY_PAIR_)
        break;
     
      Cluster c = clustering.join(best_pair.getPair().getLeft(), best_pair.getPair().getRight());
      clustering.updateFeatureVectors(doc, rm.alphabet_, c, rm);
      
      if (++decisions % 10 == 0)
        System.out.println("\t  " + decisions + " clustering decisions.");
      
      best_pair = getBestScoringPair(doc, clustering);
    }
  }
  
  
  public void train(Document[] docs)
  {
    PerceptronRankingModel rm = (PerceptronRankingModel) rm_;
    
    for (int epoch = 0; epoch < NUM_EPOCHS_; epoch++) {
      System.out.println("Training epoch " + epoch + ".");
      for (Document doc : docs) {
        System.out.println("\tDocument " + doc.id_);
        updateModelPerceptron(doc);
      }
    }

    rm.setAverageDiscriminant();
    rm.alphabet_.stopGrowth();
    rm.printFeatureWeights();
  }
  
  
  static public class DecisionInfo
  {
    private Pair<Cluster> pair_;
    
    public double score_;
    public int true_pairs_;
    public int total_pairs_;
    
    public DecisionInfo(Pair<Cluster> pair)
    {
      pair_ = pair;
      if (pair == Cluster.EMPTY_PAIR_) {
        true_pairs_ = 1;
        total_pairs_ = 2;
      }
      else {
        Cluster l = pair.getLeft();
        Cluster r = pair.getRight();
        true_pairs_ = getTruePairs(l, r);
        total_pairs_ = l.size() * r.size();
      }
    }
    
    public Pair<Cluster> getPair()
    {
      return pair_;
    }
    
    public int getTruePairs(Cluster l, Cluster r)
    {
      int count = 0;
      for (Element el1 : l) {
        for (Element el2 : r) {
          if (el1.getClusterID() == el2.getClusterID())
            count++;
        }
      }
      
      return count;
    }
    
    public int compareAccuracy(DecisionInfo decision)
    {
      int left = true_pairs_ * decision.total_pairs_;
      int right = decision.true_pairs_ * total_pairs_;
      
      if (left == right)
        return 0;
      if (left < right)
        return -1;
      else
        return 1;
    }
  }
  
  
  protected void updateModelPerceptron(Document doc)
  {
    PerceptronRankingModel rm = (PerceptronRankingModel) rm_;
    
    Clustering clustering = doc.getSingletonClusters();
    clustering.setFeatureVectors(doc, rm.alphabet_, rm);
    
    int decisions = 0;
    
    DecisionInfo best_pair = getBestScoringPair(doc, clustering); // can be empty
    while (best_pair != null) {
      DecisionInfo better_pair = getBetterAccuracyPair(doc, clustering, best_pair); // can be empty

      // Higher accuracy pair should have higher score.
      if (better_pair != null)
        rm.update(doc, Pair.create(better_pair.getPair(), best_pair.getPair()));
      rm.updateAvg();
      
      if (best_pair.getPair() == Cluster.EMPTY_PAIR_)
        break;
     
      Cluster c = clustering.join(best_pair.getPair().getLeft(), best_pair.getPair().getRight());
      clustering.updateFeatureVectors(doc, rm.alphabet_, c, rm);
      
      if (++decisions % 10 == 0)
        System.out.println("\t  " + decisions + " clustering decisions.");
      
      best_pair = getBestScoringPair(doc, clustering);
    }
  }
  
  
  public DecisionInfo getBetterAccuracyPair(Document doc, Clustering clustering, DecisionInfo pivot)
  {
    Discriminant disc = rm_.getDiscriminant();
    
    DecisionInfo best_decision = null;
    double best_score = Double.NEGATIVE_INFINITY;
    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);
        if (!doc.isAllowedPair(c1, c2))
          continue;

        // Need higher accuracy then the pivot.
        DecisionInfo current = new DecisionInfo(new Pair<Cluster>(c1, c2));
        if (current.compareAccuracy(pivot) > 0) {
          if (clustering.fvs_ != null)
            current.score_ = disc.evaluate(clustering.fvs_.get(i, j));
          else {
            // current.score_ = rm_.evaluate(doc, Pair.create(c1, c2));
            FeatureVector fv = doc.getFeatureVector(Pair.create(c1, c2), rm_.alphabet_);
            if (fv.features_.size() == 0)
              continue;
            current.score_ = rm_.getDiscriminant().evaluate(fv);
          }
          if (current.score_ > best_score + EPSILON_) {
            best_decision = current;
            best_score = current.score_;
          }
          else if (current.score_ >= best_score - EPSILON_ && current.compareAccuracy(best_decision) > 0) {
            best_decision = current;
            best_score = current.score_;
          }
        } 
      }
    }
    
    // Compare against empty decision.
    DecisionInfo empty_decision = new DecisionInfo(Cluster.EMPTY_PAIR_);
    if (empty_decision.compareAccuracy(pivot) > 0) {
      empty_decision.score_ = rm_.evaluate(doc, Cluster.EMPTY_PAIR_);
      if (empty_decision.score_ > best_score) {
        best_decision = empty_decision;
        best_score = empty_decision.score_;
      }
    }

    return best_decision;
  }

  
  public DecisionInfo getBestScoringPair(Document doc, Clustering clustering)
  {
    if (clustering.size() < 2)
      return null;
    
    Discriminant disc = rm_.getDiscriminant();
    
    Pair<Cluster> best_pair = Cluster.EMPTY_PAIR_;
    double best_score = rm_.evaluate(doc, best_pair);

    Vector<Pair<Cluster>> bpairs = new Vector<Pair<Cluster>>();
    bpairs.add(best_pair);
    
    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);
        if (!doc.isAllowedPair(c1, c2))
          continue;
        
        double score;
        if (clustering.fvs_ != null)
          score = disc.evaluate(clustering.fvs_.get(i, j));
        else {
          // score = rm_.evaluate(doc, Pair.create(c1, c2));
          FeatureVector fv = doc.getFeatureVector(Pair.create(c1, c2), rm_.alphabet_);
          if (fv.features_.size() == 0)
            continue;
          score = rm_.getDiscriminant().evaluate(fv);          
        }
        
        if (score > best_score + EPSILON_) { // find another way to kickstart, than >=
          best_pair = new Pair<Cluster>(c1, c2);
          best_score = score;
          bpairs.clear();
          bpairs.add(best_pair);
        }
        else if (score >= best_score - EPSILON_) {
          best_pair = new Pair<Cluster>(c1, c2);
          bpairs.add(best_pair);
        }
      }
    }
    
    int pos = random_.nextInt(bpairs.size());
    DecisionInfo best_decision = new DecisionInfo(bpairs.get(pos));
    best_decision.score_ = best_score;
    
    return best_decision;
  }


  public Pair<Cluster> getBestNonEmptyPair(Document doc, Clustering clustering)
  {
    Discriminant disc = rm_.getDiscriminant();
    
    Pair<Cluster> best_pair = null;
    double best_score = Double.NEGATIVE_INFINITY;
    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);
        if (!doc.isAllowedPair(c1, c2))
          continue;
         
        double score;
        if (clustering.fvs_ != null)
          score = disc.evaluate(clustering.fvs_.get(i, j));
        else {
          // score = rm_.evaluate(doc, Pair.create(c1, c2));
          FeatureVector fv = doc.getFeatureVector(Pair.create(c1, c2), rm_.alphabet_);
          if (fv.features_.size() == 0)
            continue;
          score = rm_.getDiscriminant().evaluate(fv);
        }
        if (score > best_score) {
          best_pair = new Pair<Cluster>(c1, c2);
          best_score = score;
        }
      }
    }
    
    return best_pair;
  }

  
  public Pair<Cluster> getBestPair(Document doc, Clustering clustering)
  {
    Discriminant disc = rm_.getDiscriminant();
    
    Pair<Cluster> best_pair = Cluster.EMPTY_PAIR_;
    double best_score = rm_.evaluate(doc, best_pair);

    for (int i = 0; i < clustering.clusters_.size(); i++) {
      Cluster c1 = clustering.clusters_.get(i);
      for (int j = i + 1; j < clustering.clusters_.size(); j++) {
        Cluster c2 = clustering.clusters_.get(j);
        if (!doc.isAllowedPair(c1, c2))
          continue;
        
        // Temporary distance constraint for efficiency.
        int distance = getElemDistance(c1, c2);
        if (distance > MAX_DISTANCE_)
          continue;
 
        double score;
        if (clustering.fvs_ != null)
          score = disc.evaluate(clustering.fvs_.get(i, j));
        else {
          // score = rm_.evaluate(doc, Pair.create(c1, c2));
          FeatureVector fv = doc.getFeatureVector(Pair.create(c1, c2), rm_.alphabet_);
          if (fv.features_.size() == 0)
            continue;
          score = rm_.getDiscriminant().evaluate(fv);
        }
        if (score > best_score) {
          best_pair = new Pair<Cluster>(c1, c2);
          best_score = score;
        }
      }
    }
        
    return best_pair;
  }



  public int getElemDistance(Cluster c1, Cluster c2)
  {
    // Assert elements are sorted inside each cluster.
    int first1 = c1.elements_[0].index_;
    int last1 = c1.elements_[c1.elements_.length - 1].index_;
    
    int first2 = c2.elements_[0].index_;
    int last2 = c2.elements_[c2.elements_.length - 1].index_;
    
    if (first1 < first2) {
      if (last1 < first2)
        return first2 - last1;
      return 0;
    }
    else {
      if (last2 < first1)
        return first1 - last2;
      return 0;
    }
  }
  
}
